;; Copyright 2022 by Joona "Retropikzel" Isoaho <retropikzel@iki.fi>
;; This program is free software: you can redistribute it and/or modify it under
;; the terms of the GNU Lesser General Public License as published by the Free Software
;; Foundation, either version 3 of the License, or (at your option) any later version.
;; This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
;; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
;; See the GNU Lesser General Public License for more details.
;;You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.

(define-library
  (retropikzel scgi)
  (import
    (scheme base)
    (scheme char)
    (srfi 106))
  (export scgi-start
          scgi-request-value-get)
  (begin

    (define scgi-request-value-get
      (lambda (request key)
        (let ((value-pair (assoc key request)))
          (if (or (null? value-pair)
                  (not value-pair))
            #f
            (let ((value (cdr value-pair)))
              (if (null? value)
                #f
                value))))))

    (define scgi-split-by-zero->list
      (lambda (source)
        (let ((result (list))
              (source-size (bytevector-length source)))
          (letrec ((looper
                     (lambda (index last-index key value)
                       (if (< index source-size)
                         (if (and key value)
                           (begin
                             (if (> (bytevector-length key ) 0)
                               (set! result
                                 (append
                                   result
                                   (list (cons (string->symbol (utf8->string key))
                                               (if (= (bytevector-length value) 0)
                                                 ""
                                                 (utf8->string value)))))))
                             (looper index last-index #f #f))
                           (if (= (bytevector-u8-ref source index) 0)
                             (let ((slice (bytevector-copy source last-index index)))
                               (if (not key)
                                 (looper (+ index 1) (+ index 1) slice value)
                                 (looper (+ index 1) (+ index 1) key slice)))
                             (looper (+ index 1) last-index key value)))))))
            (looper 0 0 #f #f))
          result)))

    (define scgi-query-string->list
      (lambda (query-string)
        (let* ((query-bytevector (string->utf8 (string-append query-string "==")))
               (query-bytevector-size (bytevector-length query-bytevector))
               (u8-equal (bytevector-u8-ref (string->utf8 "=") 0))
               (u8-and (bytevector-u8-ref (string->utf8 "&") 0)))
          (letrec ((looper (lambda (index)
                             (if (< index query-bytevector-size)
                               (begin
                                 (if (or (= (bytevector-u8-ref query-bytevector index) u8-equal)
                                         (= (bytevector-u8-ref query-bytevector index) u8-and))
                                   (bytevector-u8-set! query-bytevector index 0))
                                 (looper (+ index 1)))))))
            (looper 0))
          (scgi-split-by-zero->list query-bytevector))))

    (define scgi-netstring->list
      (lambda (netstring)
        (let ((request (list)))
          (letrec ((get-request
                     (lambda (index)
                       (if (= (bytevector-u8-ref netstring index) 58)
                         (bytevector-copy netstring (+ index 1))
                         (get-request (+ index 1))))))
            (if (> (bytevector-length netstring) 0)
              (scgi-split-by-zero->list (get-request 0))
              (list))))))


    (define scgi-get-request-body
      (lambda (request-bytes content-length)
        (letrec ((looper
                   (lambda (index)
                     (if (and (> (bytevector-length request-bytes) 0)
                              (= (bytevector-u8-ref request-bytes index) 0)
                              (= (bytevector-u8-ref request-bytes (+ index 1)) 44))
                       (utf8->string (bytevector-copy request-bytes (+ index 2)))
                       (looper (- index 1))))))
          (looper (- (bytevector-length request-bytes) 1)))))

    (define scgi-handle
      (lambda (client-socket handler)
        (let* ((request-bytes (socket-recv client-socket 1000000))
               (request (scgi-netstring->list request-bytes))
               (request-method (if (not (null? request)) (cdr (assoc 'REQUEST_METHOD request)) ""))
               (request-uri (if (not (null? request)) (cdr (assoc 'REQUEST_URI request)) ""))
               (content-length (if (not (null? request)) (string->number (cdr (assoc 'CONTENT_LENGTH request))) 0))
               (body (if (> content-length 0) (scgi-get-request-body request-bytes content-length) ""))
               (arguments (if (string-ci=? request-method "POST")
                            (if (not (null? request)) (scgi-query-string->list body) (list))
                            (if (not (null? request)) (scgi-query-string->list (cdr (assoc 'QUERY_STRING request))) (list)))))
          (let ((response (handler request)))
            (socket-send client-socket (string->utf8 (if (string? response)
                                                       response
                                                       "")))))
        (socket-close client-socket)))

    (define scgi-listen
      (lambda (socket handler)
        (scgi-handle (socket-accept socket) handler)
        (scgi-listen socket handler)))

    (define scgi-start
      (lambda (port handler)
        (let ((socket (make-server-socket port)))
          (scgi-listen socket handler))))))
